# _*_ coding: utf-8 _*_
#
# hspylib v0.11.200
#
# Package: main.hspylib.modules.application
"""Package initialization."""

__all__ = [
    'application', 
    'argparse', 
    'exit_hooks', 
    'exit_status', 
    'version'
]
__version__ = '0.11.200'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: hspylib
   @package: hspylib.modules.fetch
      @file: uri_scheme.py
   @created: Mon, 12 Dec 2022
    @author: "<B>H</B>ugo <B>S</B>aporetti <B>J</B>unior")"
      @site: "https://github.com/yorevs/hspylib")
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2023, HsPyLib team
"""

from hspylib.core.enums.enumeration import Enumeration


class UriScheme(Enumeration):
    """Uniform Resource Identifier helps identify a source without ambiguity
    Ref.: https://en.wikipedia.org/wiki/List_of_URI_schemes
    """

    # fmt: off
    ABOUT       = 'about'
    HTTP        = 'http'
    HTTPS       = 'https'
    FTP         = 'ftp'
    FILE        = 'file'
    # fmt: on

    @classmethod
    def of(cls, scheme: str) -> "UriScheme":
        """Create a URI scheme from a string."""
        try:
            return super().of_value(scheme, ignore_case=True)
        except TypeError as err:
            raise NotImplementedError(f"'{scheme}' scheme is not supported") from err



# _*_ coding: utf-8 _*_
#
# hspylib v0.11.200
#
# Package: main.hspylib.core
"""Package initialization."""

__all__ = [
    'collection_filter', 
    'config', 
    'constants', 
    'decorator', 
    'enums', 
    'exception', 
    'metaclass', 
    'namespace', 
    'preconditions', 
    'tools', 
    'zoned_datetime'
]
__version__ = '0.11.200'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib
   @package: hspylib.modules.application
      @file: application.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2023, HsPyLib team
"""
from hspylib.core.tools.commons import syserr
from hspylib.modules.application.exit_status import ExitStatus
from typing import Callable, Optional

import sys
import traceback

CLEANUP_CB = Optional[Callable[[], None]]

EXIT_CB = Callable[[int | ExitStatus], None]

EXIT_STATUS = Optional[ExitStatus | int]


class ExitHooks:
    """Provide application exit hooks."""

    def __init__(self, cleanup: CLEANUP_CB = None):
        self._orig_exit: EXIT_CB = sys.exit
        self._exit_status: ExitStatus = ExitStatus.SUCCESS
        self._exception = None
        self._traceback = None
        self._was_hooked: bool = False
        self._cleanup: CLEANUP_CB = cleanup
        sys.exit = self.exit

    def hook(self) -> None:
        """Set the default system exception hook to this hook."""
        sys.excepthook = self.except_hook

    def exit(self, exit_status: EXIT_STATUS) -> None:
        """Exit the application with the specified exit status."""
        self._exit_status = ExitStatus.of(int(exit_status or 0))
        if not self._was_hooked:
            self._orig_exit(int(self._exit_status))
        else:
            exit(int(self._exit_status))

    def except_hook(self, exc_type: TypeError | None, exc: BaseException | None, tb: traceback) -> None:
        """Handle a system exception."""
        self._was_hooked = True
        self._exception, self._traceback = exc, tb
        tb = "\t".join(traceback.format_exception(exc_type, exc, self._traceback))
        syserr(tb)
        if self._cleanup:
            self._cleanup()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib
   @package: hspylib.modules.cli.vt100
      @file: vt_100.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2023, HsPyLib team
"""
from abc import ABC
from hspylib.core.preconditions import check_argument

import os
import re


class Vt100(ABC):
    """
    References:
        - https://vt100.net/docs/vt100-ug/chapter3.html
        - https://espterm.github.io/docs/VT100%20escape%20codes.html
    """

    TERM = os.environ.get("TERM", "xterm-color").lower()

    # Esc<Sequence>
    @staticmethod
    def escape(seq: str) -> str:
        """Build an escape code start string."""
        return f"\033{seq}"

    # Esc[<Code>
    @staticmethod
    def sequence(code: str) -> str:
        """Build an escape sequence start string."""
        return Vt100.escape(f"[{code}")

    # Esc7
    @staticmethod
    def save_cursor() -> str:
        """Build an escape sequence for saving cursor position and attributes."""
        return Vt100.escape("7")

    # Esc8
    @staticmethod
    def restore_cursor() -> str:
        """Build an escape sequence for restoring cursor position and attributes."""
        return Vt100.escape("8")

    # Esc[c
    @staticmethod
    def reset() -> str:
        """Build an escape sequence to reset terminal to initial state."""
        return Vt100.sequence("c")

    # Esc[?7<h/l>
    @staticmethod
    def set_auto_wrap(enabled: bool) -> str:
        """Build an escape sequence to set auto-wrap mode on/off."""
        return Vt100.sequence(f"?7{'h' if enabled else 'l'}")

    # Esc[?25<h/l>
    @staticmethod
    def set_show_cursor(enabled: bool) -> str:
        """Build an escape sequence to set show-cursor mode on/off."""
        return Vt100.sequence(f"?25{'h' if enabled else 'l'}")

    # Esc[?25<h/l>
    @staticmethod
    def get_cursor_pos() -> str:
        """Build an escape sequence to get cursor position. Response: cursor is at v,h."""
        return Vt100.sequence("6n")

    # Esc[<Modes...>m
    @staticmethod
    def mode(mod_seq: str) -> str:
        """Build an escape sequence to change character attributes. 'Esc[m' or 'Esc[0m' ro reset."""
        check_argument(bool(re.match(r"[0-9]+(;[0-9]+){0,2}", mod_seq)), f"Invalid mode sequence: {mod_seq}")
        return Vt100.sequence(f"{mod_seq}m")

    # Esc[<n>J
    @staticmethod
    def clear_screen(mod_cls: int = None) -> str:
        """Build an escape sequence to clear portions or the entire screen."""
        if not mod_cls:
            return Vt100.sequence("J")
        check_argument(mod_cls in [0, 1, 2], f"Invalid clear screen sequence: {mod_cls}")
        return Vt100.sequence(f"{mod_cls}J")

    # Esc[<n>K
    @staticmethod
    def clear_line(mod_cls: int = None) -> str:
        """Build an escape sequence to clear portions or the entire cursor line."""
        if not mod_cls:
            return Vt100.sequence("K")
        check_argument(mod_cls in [0, 1, 2], f"Invalid clea line sequence: {mod_cls}")
        return Vt100.sequence(f"{mod_cls}K")

    # Esc[<v>;<h>H
    @staticmethod
    def set_cursor_pos(cup_seq: str = None) -> str:
        """Build an escape sequence to set cursor position."""
        if not cup_seq:
            return Vt100.sequence("H")
        check_argument(bool(re.match(r"[0-9]*;[0-9]*", cup_seq)), f"Invalid cursor position sequence: {cup_seq}")
        return Vt100.sequence(f"{cup_seq}H")

    # Esc[<n><A/B/C/D>
    @staticmethod
    def cursor_move(amount: int, direction: str) -> str:
        """Build an escape sequence to move the cursor to a direction."""
        check_argument(
            int(amount) >= 0 and direction in ["A", "B", "C", "D"],
            f"Invalid direction={direction} or move amount={amount}",
        )
        return Vt100.sequence(f"{amount}{direction}")

    # Esc[<n>A
    @staticmethod
    def cursor_move_up(amount: int = None) -> str:
        """Wrapper to build an escape sequence to move the cursor UP."""
        return Vt100.cursor_move(amount or 0, "A")

    # Esc[<n>B
    @staticmethod
    def cursor_move_down(amount: int = None) -> str:
        """Wrapper to build an escape sequence to move the cursor DOWN."""
        return Vt100.cursor_move(amount or 0, "B")

    # Esc[<n>C
    @staticmethod
    def cursor_move_forward(amount: int = None) -> str:
        """Wrapper to build an escape sequence to move the cursor FORWARD/RIGHT."""
        return Vt100.cursor_move(amount or 0, "C")

    # Esc[<n>D
    @staticmethod
    def cursor_move_backward(amount: int = None) -> str:
        """Wrapper to build an escape sequence to move the cursor BACKWARD/LEFT."""
        return Vt100.cursor_move(amount or 0, "D")

    # Esc[?1049<h/l>
    @staticmethod
    def alternate_screen(enable: bool = True) -> str:
        """Build an escape sequence to switch to the alternate screen."""
        seq = NotImplemented
        match Vt100.TERM:
            case "xterm-256color":
                seq = Vt100.sequence(f"?1049{'h' if enable else 'l'}")
            case "xterm-color":
                if enable:
                    seq = Vt100.escape("7") + Vt100.sequence("?47h")
                else:
                    seq = Vt100.sequence("2J") + Vt100.sequence("?47l") + Vt100.escape("8")
        return seq



# _*_ coding: utf-8 _*_
#
# hspylib v0.11.200
#
# Package: main.hspylib.core.enums
"""Package initialization."""

__all__ = [
    'charset', 
    'content_type', 
    'enumeration', 
    'http_code', 
    'http_method'
]
__version__ = '0.11.200'



# _*_ coding: utf-8 _*_
#
# hspylib v0.11.200
#
# Package: main.hspylib.modules
"""Package initialization."""

__all__ = [
    'application', 
    'cache', 
    'cli', 
    'eventbus', 
    'fetch', 
    'security'
]
__version__ = '0.11.200'



# _*_ coding: utf-8 _*_
#
# hspylib v0.11.200
#
# Package: main.hspylib.core.config
"""Package initialization."""

__all__ = [
    'app_config', 
    'parser_factory', 
    'properties'
]
__version__ = '0.11.200'



# _*_ coding: utf-8 _*_
#
# hspylib v0.11.200
#
# Package: main.hspylib.core.tools
"""Package initialization."""

__all__ = [
    'commons', 
    'dict_tools', 
    'json_path', 
    'text_tools', 
    'validator'
]
__version__ = '0.11.200'



# _*_ coding: utf-8 _*_
#
# hspylib v0.11.200
#
# Package: main.hspylib.modules.application.argparse
"""Package initialization."""

__all__ = [
    'argument_parser', 
    'arguments_builder', 
    'chained_arguments_builder', 
    'options_builder', 
    'parser_action'
]
__version__ = '0.11.200'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib
      @file: setup.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2023, HsPyLib team
"""

import pathlib
import setuptools

HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# The version of the package
VERSION = (HERE / "hspylib/.version").read_text().strip()

# The package requirements
REQUIREMENTS = list(filter(None, (HERE / "requirements.txt").read_text().splitlines()))

# This call to setup() does all the work
setuptools.setup(
    name="hspylib",
    version=VERSION,
    description="HsPyLib - Core python library",
    author="Hugo Saporetti Junior",
    author_email="yorevs@hotmail.com",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/yorevs/hspylib",
    project_urls={"GitHub": "https://github.com/yorevs/hspylib", "PyPi": "https://pypi.org/project/hspylib/"},
    license="MIT",
    license_files="LICENSE.md",
    packages=setuptools.find_namespace_packages(),
    include_package_data=True,
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.10",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Unix",
        "Development Status :: 4 - Beta",
        "Natural Language :: English",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Intended Audience :: Developers",
    ],
    python_requires=">=3.10",
    install_requires=REQUIREMENTS,
    keywords="python,library,solid,patterns,development,integration,application,framework",
    platforms="Darwin,Linux",
)



# _*_ coding: utf-8 _*_
#
# hspylib v0.11.200
#
# Package: main.hspylib.modules.cli.vt100
"""Package initialization."""

__all__ = [
    'vt_100', 
    'vt_code', 
    'vt_color'
]
__version__ = '0.11.200'



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib
   @package: hspylib.modules.application.argparse
      @file: argument_chain_builder.py
   @created: hu, 14 Feb 2022
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2023, HsPyLib team
"""
from argparse import ArgumentParser
from functools import partial
from hspylib.modules.application.argparse.parser_action import ParserAction
from typing import Any, Union


class ChainedArgumentsBuilder:
    """Class to provide chained arguments parser builder."""

    def __init__(self, arg_parser: ArgumentParser, subcommand_name: str, subcommand_help: str):
        self._arg_parser = arg_parser
        self._subparsers = self._arg_parser.add_subparsers(
            title=subcommand_name, dest=subcommand_name, help=subcommand_help, required=True
        )
        self._current = arg_parser

    def argument(self, name: str, help_string: str = None) -> "ChainedArgumentsBuilder":
        """Assign a new chained argument to the parser."""

        self._current = self._subparsers.add_parser(name, help=help_string)

        return self

    def add_parameter(
        self,
        name: str,
        help_string: str = None,
        choices: list = None,
        nargs: Union[str, int] = None,
        action: ParserAction = ParserAction.STORE,
    ) -> "ChainedArgumentsBuilder":
        """Assign a new chained argument parameter to the parser."""

        add_argument = partial(
            self._current.add_argument, help=help_string or f"the {name}", dest=name, action=action.value
        )

        if action == ParserAction.STORE:
            add_argument(choices=choices, nargs=nargs)
        else:
            add_argument()

        return self

    def add_option(
        self,
        name: str,
        shortopt: str,
        longopt: str,
        help_string: str = None,
        choices: list = None,
        required: bool = False,
        nargs: Union[str, int] = None,
        action: ParserAction = ParserAction.STORE,
        default: Any = None,
    ) -> "ChainedArgumentsBuilder":
        """Assign a new chained option parameter to the parser."""

        add_arg = partial(
            self._current.add_argument,
            f"-{shortopt.replace('^-', '')[0]}",
            f"--{longopt.replace('^-*', '')}",
            dest=name,
            help=help_string or f"the {longopt}",
            required=required,
            action=action.value,
            default=default,
        )
        if action == ParserAction.STORE:
            add_arg(choices=choices, nargs=nargs)
        else:
            add_arg()

        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
   @project: HsPyLib
   @package: hspylib.core.tools
      @file: json_path.py
   @created: Tue, 4 May 2021
    @author: <B>H</B>ugo <B>S</B>aporetti <B>J</B>unior"
      @site: https://github.com/yorevs/hspylib
   @license: MIT - Please refer to <https://opensource.org/licenses/MIT>

   Copyright 2023, HsPyLib team
"""

from pyparsing import unicode
from typing import Any, TypeVar

import re

JSON_ELEMENT = TypeVar("JSON_ELEMENT", bound=[list, dict, unicode])


class JsonPath:
    """Navigate and"""

    RE_JSON_NAME = "[a-zA-Z0-9_\\- ]"
    RE_JSON_ARRAY_INDEX = "[0-9]{1,}"

    def __init__(
        self, separator: str = ".", json_name_re: str = RE_JSON_NAME, json_array_index_re: str = RE_JSON_ARRAY_INDEX
    ):
        """Construction"""

        self.separator = separator
        self.jsonNameRe = json_name_re
        self.jsonArrayIndexRe = json_array_index_re
        self.pat_elem = None
        self.pat_sel_elem_val = None
        self.pat_sub_expr = None
        self.pat_sub_expr_val = None

    def _find_next_element(
        self, root_element: JSON_ELEMENT, match_name: str, match_value: Any = None, fetch_parent: bool = False
    ) -> JSON_ELEMENT:
        """Find the next element in the list matching the specified value."""

        selected_element = root_element
        if isinstance(selected_element, list):
            for nextInList in root_element:
                if isinstance(nextInList, dict):
                    selected_element = nextInList.get(match_name)
                    if selected_element and (match_value is None or (match_value and selected_element == match_value)):
                        # To return the parent element instead of the leaf
                        if fetch_parent:
                            selected_element = nextInList
                        break
                    selected_element = None
                elif isinstance(nextInList, unicode):
                    selected_element = root_element
                elif isinstance(nextInList, list):
                    selected_element = self._find_next_element(nextInList, match_name, match_value)
        elif isinstance(selected_element, dict):
            el = selected_element.get(match_name)
            if el and (match_value is None or (match_value and el == match_value)):
                # To return the parent element instead of the leaf
                if not fetch_parent:
                    selected_element = el
            else:
                selected_element = None
        else:
            selected_element = None

        return selected_element

    def _find_in_sub_expr(
        self, sub_expr: JSON_ELEMENT, sub_sel_el: JSON_ELEMENT, pat_subst_expr_val: Any, fetch_parent: bool = False
    ) -> JSON_ELEMENT:
        """Find the element in the sub-expressions."""

        for nextSubExpr in sub_expr:
            if nextSubExpr:
                sub_parts = re.search(pat_subst_expr_val, nextSubExpr)
                sub_elem_id = sub_parts.group(1)
                sub_elem_val = sub_parts.group(3)
                sub_sel_el = self._find_next_element(sub_sel_el, sub_elem_id, sub_elem_val, fetch_parent)

        return sub_sel_el

    # pylint: disable=too-many-branches,consider-using-f-string
    def select(self, root_element: JSON_ELEMENT, search_path: str, fetch_parent: bool = False) -> JSON_ELEMENT:
        """
        Get the json element through it's path. Returned object is either [dict, list or unicode].

        Search patterns:
          1. elem1.elem2
          2. elem1.elem2[index]
          3. elem1.elem2{property}
          4. elem1.elem2{property}[index]
          5. elem1.elem2{property<value>}
          6. elem1.elem2[index].elem3
          7. elem1.elem2{property}.elem3
          8. elem1.elem2{property<value>}.elem3
          9. elem1.elem2{property<value>}[index].elem3
         10. elem1.elem2{property<value>}.{property2<value2>}.elem3
        """

        self.pat_elem = f"{self.jsonNameRe}+"
        self.pat_sel_elem_val = "(%s)?((\\{(%s)(<(%s)>)?\\})+)(\\[(%s)\\])?" % (
            self.pat_elem,
            self.pat_elem,
            self.pat_elem,
            self.jsonArrayIndexRe,
        )
        self.pat_sub_expr = "(\\{%s\\})" % self.pat_elem
        self.pat_sub_expr_val = "\\{(%s)(<(%s)>)?\\}" % (self.pat_elem, self.pat_elem)
        selected_element = root_element

        # pylint: disable=too-many-nested-blocks
        try:
            search_tokens = search_path.split(self.separator)
            for nextElement in search_tokens:
                if nextElement.find("{") >= 0:  # Next element has nested elements
                    parts = re.search(self.pat_sel_elem_val, nextElement)
                    sel_elem_id = parts.group(1)
                    sub_parts = parts.group(2)
                    elem_array_group = parts.group(7)
                    elem_array_index = parts.group(8)
                    sub_expressions = re.compile(self.pat_sub_expr).split(sub_parts)
                    if sel_elem_id and isinstance(selected_element, dict):
                        selected_element = selected_element.get(sel_elem_id)
                    if selected_element:
                        # Our first element is a list, so we will have to loop and find all the elements
                        # and sub expressions in it.
                        if isinstance(selected_element, list):
                            for nextInList in selected_element:
                                sub_selected_element = self._find_in_sub_expr(
                                    sub_expressions, nextInList, self.pat_sub_expr_val, fetch_parent
                                )
                                # It sub_selected_element is not null then we have found what we wanted.
                                if sub_selected_element:
                                    selected_element = sub_selected_element
                                    break
                        # Check if there are indexed elements.
                        if elem_array_group and elem_array_index and isinstance(selected_element, list):
                            selected_element = selected_element[int(elem_array_index)]

                elif nextElement.find("[") >= 0:  # Next element is indexed
                    pat_sel_elem_idx = f"({self.pat_elem})\\[({self.jsonArrayIndexRe})\\]"
                    parts = re.search(pat_sel_elem_idx, nextElement)
                    sub_elem_id = parts.group(1)
                    elem_array_index = parts.group(2)
                    # TODO Implement subarray like elem[0][1][2]
                    if sub_elem_id is not None and elem_array_index is not None:
                        el = selected_element.get(sub_elem_id)
                        if isinstance(el, list):
                            selected_element = el[int(elem_array_index)]
                else:  # Next element is simple
                    selected_element = selected_element.get(nextElement)
        except (AttributeError, IndexError):
            selected_element = None

        return selected_element
